/* upsfetch - library for UPS communications from client programs

   Copyright (C) 1999  Russell Kroll <rkroll@exploits.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              
*/

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/socket.h>
#include "config.h"
#include "version.h"

	int	upserror, syserrno;

#define UNKNOWN		0x1000
#define NOSUCHVAR 	0x1001
#define NOSUCHHOST	0x1002
#define SENDFAILURE	0x1003
#define RECVFAILURE	0x1004
#define SOCKFAILURE	0x1005
#define BINDFAILURE	0x1006
#define RECVTIMEOUT	0x1007
#define	NOCOMM1		0x1008
#define UNKNOWNUPS	0x1009
#define ACCESSDENIED	0x100A
#define CONNFAILURE	0x100B
#define READERROR	0x100C
#define WRITEERROR	0x100D

char *upsstrerror (int errnum)
{
	char	buf[256];

	switch (errnum) {
		case UNKNOWN:
			return (strdup("Unknown error"));
			break;
		case NOSUCHVAR: 
			return (strdup("No such variable"));
			break;
		case NOSUCHHOST:
			return (strdup("No such host"));
			break;
		case SENDFAILURE:
			snprintf (buf, sizeof(buf), "Send failure: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		case RECVFAILURE:
			snprintf (buf, sizeof(buf), "Receive failure: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		case SOCKFAILURE:
			snprintf (buf, sizeof(buf), "Socket failure: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		case BINDFAILURE:
			snprintf (buf, sizeof(buf), "bind failure: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		case RECVTIMEOUT:
			return (strdup("Receive timeout"));
			break;
		case NOCOMM1:
			return (strdup("Data source is stale"));
			break;
		case UNKNOWNUPS:
			return (strdup("Unknown ups name"));
			break;
		case ACCESSDENIED:
			return (strdup("Access denied"));
			break;
		case CONNFAILURE:
			snprintf (buf, sizeof(buf), "Connection failure: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		case READERROR:
			snprintf (buf, sizeof(buf), "Read error: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		case WRITEERROR:
			snprintf (buf, sizeof(buf), "Write error: %s",
			         strerror(syserrno));
			return (strdup(buf));
			break;
		default:
			snprintf (buf, sizeof(buf), "Unknown error 0x%04x", errnum);
			return (strdup(buf));
			break;
	}

	/* notreached */
	return (NULL);
}

/* common error condition handler */
int errcheck (char *buf)
{
	if (!strncmp(buf, "ERR", 3)) {
		upserror = UNKNOWN;		/* default error */

		if (!strncmp(buf, "ERR ACCESS-DENIED", 17)) 
			upserror = ACCESSDENIED;

		if (!strncmp(buf, "ERR UNKNOWN-UPS", 15))
			upserror = UNKNOWNUPS;

		return (-1);
	}

	return 0;
}

/* send <req> to <host> and put the answer in <buf> */

int fetch (char *hostin, char *req, char *buf, int buflen)
{
	int	res, udpport = UDPPORT, fd, fromlen, err;
	struct	sockaddr_in xmit, dest, from;
	struct	hostent *serv;
	char	sbuf[512], host[512], *cpos;
	fd_set	rfd;
	struct	timeval tv;

	upserror = UNKNOWN;

	snprintf (host, sizeof(host), "%s", hostin);
	cpos = strstr (host, ":");

	if (cpos != NULL) {
		cpos[0] = 0;
		udpport = atoi(++cpos);
	}

	if ((serv = gethostbyname(host)) == (struct hostent *) NULL) {
		upserror = NOSUCHHOST;
		return (-1);
	}

	fd = socket (AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		upserror = SOCKFAILURE;
		syserrno = errno;
		return (-1);
	}

	memset (&xmit, '\0', sizeof(xmit));
	xmit.sin_family = AF_INET;

	res = bind (fd, (struct sockaddr *) &xmit, sizeof(xmit));
	if (res < 0) {
		upserror = BINDFAILURE;
		syserrno = errno;
		close (fd);
		return (-1);
	}

	memset (&dest, '\0', sizeof(dest));
	memcpy (&dest.sin_addr, serv->h_addr, serv->h_length);
	dest.sin_family = AF_INET;
	dest.sin_port = htons(udpport);

	snprintf (sbuf, sizeof(sbuf), "%s", req);
	res = sendto (fd, sbuf, strlen(sbuf), 0, 
	             (struct sockaddr *) &dest, sizeof(dest));

	if (res < 0) {
		close (fd);
		syserrno = errno;
		upserror = SENDFAILURE;
		return (-1);	/* failure */
	}

	FD_ZERO (&rfd);
	FD_SET (fd, &rfd);
	tv.tv_sec = 2;	/* 2.0 seconds allowed for reply */
	tv.tv_usec = 0;

	res = select (fd + 1, &rfd, NULL, NULL, &tv);

	if (res > 0) {
		memset (buf, '\0', buflen);
		fromlen = sizeof(from);
		res = recvfrom (fd, buf, buflen, 0, (struct sockaddr *) &from,
		                &fromlen);
	}
	else {
		close (fd);
		syserrno = errno;
		upserror = RECVTIMEOUT;
		return (-1);	/* failure */
	}

	if (res < 0) {
		close (fd);
		syserrno = errno;
		upserror = RECVFAILURE;
		return (-1);	/* failure */
	}

	close (fd);

	if ((err = errcheck (buf)) != 0)
		return (err);

	upserror = 0;
	return (1);	/* success */
}

int getupsvar (char *hostin, char *varname, char *buf, int buflen)
{
	char	cmd[256], tmp[512], *ptr, *host, *upsname, hosttmp[512];
	int	ret;

	/* host = ups-1@localhost */

	snprintf (hosttmp, sizeof(hosttmp), "%s", hostin);
	ptr = strstr (hosttmp, "@");

	if (ptr != NULL) {
		ptr[0] = 0;
		upsname = hosttmp;
		host = ptr + 1;
		snprintf (cmd, sizeof(cmd), "REQ %s@%s", varname, upsname);
	}
	else {
		upsname = NULL;
		host = hosttmp;
		snprintf (cmd, sizeof(cmd), "REQ %s", varname);
	}

	ret = fetch (host, cmd, tmp, sizeof(tmp));

	if (ret != 1) {
		return (ret);
	}

	tmp[strlen(tmp) - 1] = 0;
	/* ANS <varname> <answer> */

	if (upsname == NULL)		/* skip "ANS <varname> " */
		ptr = tmp + strlen(varname) + 5;
	else				/* skip "ANS <varname>@<upsname> " */
		ptr = tmp + strlen(varname) + 5 + strlen(upsname) + 1;

	if (!strcmp(ptr, "NOT-SUPPORTED")) {
		upserror = NOSUCHVAR;
		return (-1);	/* failure */
	}

	if (!strcmp(ptr, "DATA-STALE")) {
		upserror = NOCOMM1;
		return (-1);	/* failure */
	}

	if (!strcmp(ptr, "UNKNOWN-UPS")) {
		upserror = UNKNOWNUPS;
		return (-1);	/* failure */
	}

	if (!strcmp(ptr, "UNKNOWN")) {
		upserror = NOSUCHVAR;
		return (-1);	/* failure */
	}

	memset (buf, '\0', buflen);
	strncpy (buf, ptr, buflen);

	return (ret);
}

int getupsvarlist (char *hostin, char *buf, int buflen)
{
	char	tmp[512], *ptr, *upsname, *host, cmd[128], hosttmp[512];
	int	ret;

	/* host = ups-1@localhost */

	snprintf (hosttmp, sizeof(hosttmp), "%s", hostin);
	ptr = strstr (hosttmp, "@");

	if (ptr != NULL) {
		ptr[0] = 0;
		host = ptr + 1;
		upsname = hosttmp;
		snprintf (cmd, sizeof(cmd), "LISTVARS %s", upsname);
	}
	else {
		upsname = NULL;
		host = hosttmp;
		snprintf (cmd, sizeof(cmd), "LISTVARS");
	}

	ret = fetch (host, cmd, tmp, sizeof(tmp));

	if (ret != 1)
		return (ret);

	/* VARS <varlist> */
	tmp[strlen(tmp) - 1] = 0;

	if (upsname == NULL)
		ptr = tmp + 5;	/* skip "VARS " */
	else
		ptr = tmp + 5 + strlen(upsname) + 2;  /* skip "VARS <name> " */

	memset (buf, '\0', buflen);
	strncpy (buf, ptr, buflen);

	return (ret);
}

int upsconnect (char *hostin)
{
	struct	sockaddr_in local, server;
	struct	hostent *serv;
	int	fd, port = UDPPORT;
	char	*cpos, *host;

	host = strdup (hostin);
	cpos = strstr (host, ":");
	if (cpos != NULL) {
		cpos[0] = 0;
		port = atoi(++cpos);
	}

	if ((serv = gethostbyname(host)) == (struct hostent *) NULL) {
		upserror = NOSUCHHOST;
		return (-1);
	}

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		upserror = SOCKFAILURE;
		syserrno = errno;
		return (-1);
	}		

	memset (&local, '\0', sizeof(struct sockaddr_in));
	local.sin_family = AF_INET;
	local.sin_port  = htons (INADDR_ANY);

	memset (&server, '\0', sizeof(struct sockaddr_in));
	server.sin_family = AF_INET;

	server.sin_port = htons (port);
	memcpy (&server.sin_addr, serv->h_addr, serv->h_length);

	if (bind (fd, (struct sockaddr *) &local, sizeof(struct sockaddr_in)) == -1) {
		upserror = BINDFAILURE;
		syserrno = errno;
		return (-1);
	}

	if (connect (fd, (struct sockaddr *) &server, sizeof(struct sockaddr_in)) == -1) {
		upserror = CONNFAILURE;
		syserrno = errno;
		return (-1);
	}

	return (fd);
}

/* read from <fd> until a LF */
int ups_readline (int fd, char *tmp, int buflen)
{
	int	tmpptr = 0, ret;
	char	ch;

	for (;;) {
		ret = read (fd, &ch, 1);

		if (ret < 1) {
			upserror = READERROR;
			syserrno = errno;
			return (ret);
		}

		tmp[tmpptr++] = ch;
		
		if ((ch == 10) || (tmpptr >= buflen))
			return 0;
	}

	return 0;
}

int getupsvarlistfd (int fd, char *upsname, char *buf, int buflen)
{
	char	tmp[512], *ptr, cmd[128];
	int	ret, err;

	if (upsname == NULL)
		snprintf (cmd, sizeof(cmd), "LISTVARS\n");
	else
		snprintf (cmd, sizeof(cmd), "LISTVARS %s\n", upsname);

	ret = write (fd, cmd, strlen(cmd));

	if (ret < 1) {
		upserror = WRITEERROR;
		syserrno = errno;
		return (-1);
	}

	memset (tmp, '\0', sizeof(tmp));

	if ((err = ups_readline (fd, tmp, buflen)) != 0)
		return (err);

	if ((err = errcheck (tmp)) != 0)
		return (err);

	/* VARS <varlist> */
	tmp[strlen(tmp) - 1] = 0;

	if (upsname == NULL)
		ptr = tmp + 5;	/* skip "VARS " */
	else
		ptr = tmp + 5 + strlen(upsname) + 2;  /* skip "VARS <name> " */

	memset (buf, '\0', buflen);
	strncpy (buf, ptr, buflen);

	return (ret);
}

int getupsvarfd (int fd, char *upsname, char *varname, char *buf, int buflen)
{
	char	cmd[256], tmp[512], *ptr;
	int	ret, err;

	if (upsname == NULL)
		snprintf (cmd, sizeof(cmd), "REQ %s\n", varname);
	else
		snprintf (cmd, sizeof(cmd), "REQ %s@%s\n", varname, upsname);

	ret = write (fd, cmd, strlen(cmd));

	if (ret < 1) {
		upserror = WRITEERROR;
		syserrno = errno;
		return (-1);
	}

	memset (tmp, '\0', sizeof(tmp));

	if ((err = ups_readline (fd, tmp, buflen)) != 0)
		return (err);

	if ((err = errcheck (tmp)) != 0)
		return (err);

	tmp[strlen(tmp) - 1] = 0;
	/* ANS <varname> <answer> */

	if (upsname == NULL)		/* skip "ANS <varname> " */
		ptr = tmp + strlen(varname) + 5;
	else				/* skip "ANS <varname>@<upsname> " */
		ptr = tmp + strlen(varname) + 5 + strlen(upsname) + 1;

	if (!strcmp(ptr, "NOT-SUPPORTED")) {
		upserror = NOSUCHVAR;
		return (-1);	/* failure */
	}

	if (!strcmp(ptr, "DATA-STALE")) {
		upserror = NOCOMM1;
		return (-1);	/* failure */
	}

	if (!strcmp(ptr, "UNKNOWN-UPS")) {
		upserror = UNKNOWNUPS;
		return (-1);	/* failure */
	}

	memset (buf, '\0', buflen);
	strncpy (buf, ptr, buflen);

	return (ret);
}

int upslogin (int fd, char *upsname, char *password)
{
	char	cmd[256], tmp[256];
	int	err, ret;

	/* send the password */

	if ((password == NULL) || (strlen(password) == 0))
		snprintf (cmd, sizeof(cmd), "PASSWORD\n");
	else
		snprintf (cmd, sizeof(cmd), "PASSWORD %s\n", password);

	ret = write (fd, cmd, strlen(cmd));

	if (ret < 1) {
		upserror = WRITEERROR;
		syserrno = errno;
		return (-1);
	}

	memset (tmp, '\0', sizeof(tmp));

	if ((err = ups_readline (fd, tmp, sizeof(tmp))) != 0)
		return (err);

	if ((err = errcheck (tmp)) != 0)
		return (err);

	tmp[strlen(tmp) - 1] = 0;

	if (strcmp(tmp, "OK") != 0) {
		/* something happened but it's not OK or ERR */
		upserror = UNKNOWN;
		return (-1);	/* failure */
	}

	/* now attempt login */

	if ((upsname == NULL) || (strlen(upsname) == 0))
		snprintf (cmd, sizeof(cmd), "LOGIN\n");
	else
		snprintf (cmd, sizeof(cmd), "LOGIN %s\n", upsname);

	ret = write (fd, cmd, strlen(cmd));

	if (ret < 1) {
		upserror = WRITEERROR;
		syserrno = errno;
		return (-1);
	}

	memset (tmp, '\0', sizeof(tmp));

	if ((err = ups_readline (fd, tmp, sizeof(tmp))) != 0)
		return (err);

	if ((err = errcheck (tmp)) != 0)
		return (err);

	tmp[strlen(tmp) - 1] = 0;

	if (!strcmp(tmp, "OK"))
		return (0);	/* success */

	/* something happened but it's not OK or ERR */
	upserror = UNKNOWN;
	return (-1);	/* failure */
}
